from typing import Tuple, Union
from PySide6.QtCore import QMargins, Qt
from PySide6.QtWidgets import QFrame, QLayout, QWidget

from ..enums.events import EventTypeVar
from ..enums.signals import SignalTypeVar
from ..core.base import WidgetBase


class Container(WidgetBase, QFrame):
    def __init__(self, 
                 child: Union[QWidget, QLayout] = None,
                 padding: Union[QMargins, Tuple[int, int, int, int]] = None,
                 uid: str = None, 
                 signals: SignalTypeVar = None, 
                 events: EventTypeVar = None, 
                 qss: str | dict = None, 
                 attrs: list[Qt.WidgetAttribute] | dict[Qt.WidgetAttribute, bool] = None, 
                 flags: list[Qt.WindowType] | dict[Qt.WindowType, bool] = None, 
                 **kwargs):
        QFrame.__init__(self)
        super().__init__(uid, signals, events, qss, attrs, flags, **kwargs)

        self.setFrameShape(QFrame.Shape.NoFrame)
        self.setFrameShadow(QFrame.Shadow.Raised)

        if child is not None:
            if isinstance(child, QWidget):
                child.setParent(self)
                if padding is not None:
                    self.setContentsMargins(*padding
                                            if isinstance(padding, (tuple, list))
                                            else padding)

            elif isinstance(child, QLayout):
                child.setProperty("parent", self)
                self.setLayout(child)
                if padding is not None:
                    self.setContentsMargins(*padding 
                                            if isinstance(padding, (tuple, list)) 
                                            else padding)
            else:
                raise ValueError("the 'child' argument must be one of the QWidget or QLayout instance.")


class Frame(Container):
    pass


from enum import Enum, auto
from darkdetect import theme


class Theme(Enum):
    dark = auto()
    light = auto()
    system = auto()

    @staticmethod
    def system_theme():
        return Theme(theme().lower())


#!/usr/bin/python3

from PySide6.QtGui import QPaintEvent
from PySide6.QtWidgets import QWidget
from ..core.base import BehaviorDeclarative


class Painter(BehaviorDeclarative):
	def __init__(self, child: QWidget) -> QWidget:
		self._parent = child
		self._parent.paintEvent = self._paint
		return self._parent

	def update(self):
		self._repaint()

	def _repaint(self):
		self._parent.update()

	def _paint(self, event: QPaintEvent):
		raise NotImplementedError


from PySide6.QtWidgets import QHBoxLayout, QWidget, QLayout
from PySide6.QtCore import Qt
from typing import Union
from .item_wrapper import HLayoutItemWrapper
from ...core.base import ObjectBase


class HLayout(ObjectBase, QHBoxLayout):
    def __init__(self,
                 children: list[Union[QWidget, QLayout, HLayoutItemWrapper]] = None,
                 **kwargs):
        QHBoxLayout.__init__(self)
        super().__init__(**kwargs)

        self._set_children(children)

    def _set_children(self, children: list[Union[QWidget, QLayout, 
                                                 HLayoutItemWrapper]]) -> None:
        if not children:
            return

        for child in children:
            if isinstance(child, QWidget):
                self.addWidget(child)

            elif isinstance(child, QLayout):
                self.addLayout(child)
            
            elif isinstance(child, HLayoutItemWrapper):
                if isinstance(child.child, QWidget):
                    self.addWidget(*child._yield_attr())

                elif isinstance(child.child, QLayout):
                    self.addLayout(*child._yield_attr())

    @staticmethod
    def wrapper(
            child: QWidget | QLayout,
            stretch: int = None,
            alignment: Qt.AlignmentFlag = None
        ) -> HLayoutItemWrapper:
        return HLayoutItemWrapper(
            child,
            stretch,
            alignment
        )


from ..enums.widgets import Widgets, WidgetTypeVar
from ..enums.signals import Signals, SignalTypeVar
from ..enums.events import Events, EventTypeVar
from .qstyle_sheet import QStyleSheet
from typing import Any, Callable, Union
from PySide6.QtCore import Qt, QObject
from PySide6.QtWidgets import QApplication, QGraphicsEffect
import caseconverter


class AbstractBase:
    """
    this class is abstract for any widget.
    """
    def __init__(self, 
                 uid: str = None,
                 signals: SignalTypeVar = None,
                 events: EventTypeVar = None,
                 **kwargs):

        self._set_uid(uid)
        self._set_events(events)
        self._set_signals(signals)
        self._set_property_from_kwargs(**kwargs)

    def _getattr(self, name: str, default: object = None) -> object:
        return getattr(self, name, default)

    def _set_uid(self, uid: str):
        if uid is not None:
            self.setObjectName(uid.strip())

    def _set_events(self, events: EventTypeVar):
        if not events:
            return

        for (event, slot) in events:
            try:
                ename = (caseconverter.camelcase(event.name)
                         if isinstance(event, Events)
                         else event.strip())

                ## self.{event} = newEventMethod
                self.__setattr__(ename, slot)
            except Exception as err:
                print("Error-Event: ", err)

    def _set_signals(self, 
                     signals: SignalTypeVar, 
                     disconnect: bool = False):
        if not signals:
            return

        for (signal, slot) in signals:
            try:
                sname = (caseconverter.camelcase(signal.name)
                         if isinstance(signal, Signals)
                         else signal.strip())

                if (attr := self._getattr(sname)) is not None:
                    if disconnect:
                        attr.disconnect(slot)
                    else:
                        attr.connect(slot)
                else:
                    ## print Warrning slot dose't found.
                    ...
            except Exception as err:
                print("Err-Signal: ", err)

    def _get_wtype(self, w_type: WidgetTypeVar = Widgets.any):
        return (w_type.value
                if isinstance(w_type, Widgets) 
                else w_type)

    def get(self, 
            uid: str, 
            w_type: WidgetTypeVar = Widgets.any) -> QObject:
        return self.findChild(self._get_wtype(w_type), 
                              uid.strip(), 
                              Qt.FindChildOption.FindChildrenRecursively)

    def _set_property_from_kwargs(self, **kwargs):
        for (name, value) in kwargs.items():
            if hasattr(self, name) or self.property(name) is not None:
                snake_case = caseconverter.snakecase(name)

                # handle signals
                if snake_case in Signals._member_names_:
                    self._getattr(name).connect(value)

                # handle events
                elif snake_case in Events._member_names_:
                    self.__setattr__(name, value)

                # handle parent parametar
                elif name == "parent":
                    if isinstance(value, str):
                        for parent in QApplication.topLevelWidgets():
                            if parent.objectName().strip() == value.strip():
                                self.setParent(parent)
                                break

                            elif (widget := parent.findChild(QObject, 
                                                             value.strip(), 
                                                             Qt.FindChildOption.FindChildrenRecursively)) is not None:
                                self.setParent(widget)
                                break

                    elif value.__class__.__name__ in (
                        "builtin_function_or_method",
                        "method_descriptor",
                        "function"
                    ):
                        self.setParent(value())
                    else:
                        self.setParent(value)

                # handle properties
                elif self.property(name) is not None:
                    self.setProperty(name, value)

                # handle callables
                elif (func := self._getattr(name)) is not None:
                    if func.__class__.__name__ in (
                        "builtin_function_or_method",
                        "method_descriptor",
                        "function"
                    ):
                        func(value)



class ObjectBase(AbstractBase):
    pass

class WidgetBase(AbstractBase):
    def __init__(self, 
                 uid: str = None, 
                 signals: SignalTypeVar = None, 
                 events: EventTypeVar = None,
                 qss: Union[str, dict, QStyleSheet] = None,
                 attrs: Union[list[Qt.WidgetAttribute], 
                              dict[Qt.WidgetAttribute, bool]] = None,
                 flags: Union[list[Qt.WindowType], 
                              dict[Qt.WindowType, bool]] = None,
                 effect: QGraphicsEffect = None,
                 **kwargs):
        super().__init__(uid, signals, events, **kwargs)

        self.setUpdatesEnabled(True)
        self.setMouseTracking(True)

        if effect is not None:
            self._set_effect(effect)

        if qss is not None:
            self._set_qss(qss)

        if attrs is not None:
            self._set_attrs(attrs)

        if flags is not None:
            self._set_flags(flags)

    def _set_effect(self, effect: QGraphicsEffect):
        effect.setProperty("parent", self)
        effect.setParent(self)
        self.setGraphicsEffect(effect)

    def _set_flags(self, flags):
        if isinstance(flags, dict):
            for flag, on in flags.items():
                self.setWindowFlag(flag, on)
        else:
            for flag in flags:
                self.setWindowFlag(flag)

    def _set_attrs(self, attrs: dict):
        if isinstance(attrs, dict):
            for attr, on in attrs.items():
                self.setAttribute(attr, on)
        else:
            for attr in attrs:
                self.setAttribute(attr)

    def _set_qss(self, qss: Union[str, dict, QStyleSheet]):
        self.qss = (QStyleSheet(qss)
                    if isinstance(qss, (str, dict)) 
                    else qss)
        self.qss._set_parent(self)
        self.qss._set_qss(self.qss._qss)


class Return:
    def __init__(self, 
                 root: Any, 
                 child: Any):

        self._root = root
        self._child = child

    @property
    def root(self):
        return self._root

    @property
    def child(self):
        return self._child


class NoneCheck:
    '''
    e.g
    NoneCheck(icon, self.setIcon)
    
    # the same!
    if icon is not None:
        self.setIcon()
    '''

    def __new__(cls, *args, **kwargs) -> Any:
        instance = super().__new__(cls)
        return instance.__init__(*args, **kwargs)

    def __init__(self, 
                 _callable: Callable,
                 value: Any) -> Any:

        if value is not None:
            return _callable(value)



## Declarative

class DuplicateKeyError(Exception):
    def __init__(self, uid: str):
        super().__init__(uid)
        print(f"'{uid}' was registered in this dictionary!")


class _TrackingDeclarative:
    def __init__(self):
        self._instances_dict = dict()

    def _register(self, uid: str, obj: Any) -> None:
        if not isinstance(uid, str):
            raise ValueError("invalid uid, must be str type.")

        if uid in self._instances_dict:
            raise DuplicateKeyError(uid)
        self._instances_dict[uid] = obj

    def _deregister(self, uid: str) -> Any:
        """ deregister widget from manager """
        if not isinstance(uid, str):
            raise ValueError("invalid uid, must be str type.")

        if uid not in self._instances_dict:
            return

        return self._instances_dict.pop(uid)

    def pop(self, uid: str) -> Any:
       self._deregister(uid)

    def get(self, uid: str) -> Any:
        return self._instances_dict.get(uid)

    def items(self) -> Any:
        return self._instances_dict.items()

TrackingDeclarative = _TrackingDeclarative()


class BehaviorDeclarative:
    """
    `__init__` method can now return a value

    ### usage example
    ```python
    class Object(BehaviorDeclarative):
        def __init__(self, *args, **kwargs):
            return ...
    ```
    
    #### NOTE: use `uid` parameter to store this class in `TrackingDeclarative`, \
        and call it with Api.fetch
    """

    def __new__(cls, *args, **kwargs) -> Any:
        instance = super().__new__(cls)
        _uid = (kwargs.pop
                if 'uid' not in
                instance.__init__.__code__.co_varnames
                else kwargs.get)("uid", None)

        if _uid is not None:
            TrackingDeclarative._register(_uid, instance)

        return instance.__init__(*args, **kwargs)


class ObjectDeclarative(ObjectBase, BehaviorDeclarative):
    pass


class WidgetDeclarative(WidgetBase, BehaviorDeclarative):
    pass



#!/usr/bin/python3

from PySide6.QtGui import QBrush, QColor, QPaintEvent, QPainter, QPen
from PySide6.QtWidgets import QWidget
from PySide6.QtCore import QRect
from ..painter import Painter


class PaintStatusEdge(Painter):
    def __init__(self, 
            child: QWidget, 
            status_color: QColor = None,
            border_color: QColor = None,
            width: int = None) -> QWidget:

        self._status_color = (
            status_color 
            if status_color is not None 
            else QColor("white")
        )

        self._border_color = (
            border_color 
            if border_color is not None 
            else QColor("#151617")
        )

        self._width = (
            width if width is not None
            else 13
        )

        return super().__init__(child)

    def set_status_color(self, color: QColor):
        self._status_color = color
        self.update()

    def _paint(self, event: QPaintEvent):
        self._parent.__class__.paintEvent(self._parent, event)

        painter = QPainter(self._parent)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing)

        pen = QPen()
        pen.setWidth(3)
        pen.setColor(self._border_color)
        painter.setPen(pen)

        # BRUSH/STATUS COLOR
        painter.setBrush(QBrush(QColor(self._status_color)))

        # DRAW
        rect = QRect(
                self._parent.width() - self._width,
                self._parent.height() - self._width,
                self._width,
                self._width
            )
        painter.drawEllipse(rect)


from PySide6.QtWidgets import QStackedWidget, QWidget
from ...core.base import WidgetBase
from ...enums.events import EventTypeVar
from ...enums.signals import SignalTypeVar
from typing import Union


class Routes:
    def __init__(self, 
                 stacked_widget: QStackedWidget = None,
                 index: str = "/") -> None:

        self._stacked_widget = stacked_widget
        self._index = index
        self._routes: dict = {}

    def add(self, route: str, widget: QWidget) -> None:
        self._routes[route] = self._stacked_widget.addWidget(widget)

    def push(self, route: str) -> None:
        self._stacked_widget.setCurrentIndex(self._routes[route])

    def pop(self) -> None:
        self._stacked_widget.setCurrentIndex(self._routes[self._index])

    def remove(self, route: str) -> None:
        index = self._routes.pop(route)
        self._stacked_widget.removeWidget(self.widget(index))

    def stackedWidget(self) -> QStackedWidget:
        return self._stacked_widget


class StackedWidget(WidgetBase, QStackedWidget):
    def __init__(self, 
                 uid: str = None,
                 children: Union[list[QWidget], dict[str, QWidget]] = None,
                 signals: SignalTypeVar = None,
                 events: EventTypeVar = None,
                 **kwargs):
        QStackedWidget.__init__(self)
        super().__init__(uid, signals, events, **kwargs)

        if isinstance(children, dict):
            self._routes = Routes(self, "/")
            self.__setattr__("route", self._routes)
            self.__setattr__("push", lambda route: self._routes.push(route))

        if children is not None:
            self._set_children(children)

    def _set_children(self, children) -> None:
        if not isinstance(children, dict):
            for child in children:
                self.addWidget(child)
        else:
            for route, child in children.items():
                self._routes.add(route, child)


#!/usr/bin/python3

from sys import argv

from PySide6.QtGui import QFont, QIcon, QPixmap
from PySide6.QtWidgets import QGridLayout, QWidget, QApplication, QLabel, QDialog, QWidget
from PySide6.QtCore import QSize, QTimer, Qt

from ...enums.events import EventTypeVar
from ...enums.signals import SignalTypeVar
from ...core.base import WidgetBase


class _SilentTextDialog(QDialog):
    def __init__(self, parent = None, close_on_mouse_press: bool = True) -> None:
        QDialog.__init__(self, parent)

        self._close_on_mouse_press = close_on_mouse_press

        self.setWindowFlags(self.windowFlags() | Qt.WindowType.FramelessWindowHint)
        self.setAttribute(Qt.WidgetAttribute.WA_TranslucentBackground, True)
        self.setAttribute(Qt.WidgetAttribute.WA_AlwaysStackOnTop, True)
        self.setWindowOpacity(0.9)

        self.setStyleSheet("""
        QDialog, QWidget {
            padding: 6px;
            border-radius: 12px;
            background-color: #0d0d0d;
            color: #ffffff;
        }
        """)

        ## Create Qt Widgetes
        self.form = QWidget(self)
        self.form.setAttribute(Qt.WidgetAttribute.WA_TranslucentBackground, True)

        self.grid_layout1 = QGridLayout(self)
        self.grid_layout2 = QGridLayout(self.form)

        self.icon = QLabel(self.form)
        self.text = QLabel(self.form)

        font = QFont()
        font.setBold(True)
        font.setPixelSize(16)

        self.text.setFont(font)
        self.text.setAlignment(Qt.AlignmentFlag.AlignCenter)
        # self.text.setWordWrap(True)

        self.grid_layout2.addWidget(self.icon, 0, 0, 1, 1, Qt.AlignmentFlag.AlignLeft)
        self.grid_layout2.addWidget(self.text, 0, 1, 1, 1, Qt.AlignmentFlag.AlignCenter)

        self.form.setLayout(self.grid_layout2)
        self.grid_layout1.addWidget(self.form)
        self.setLayout(self.grid_layout1)

    def start_closeing(self):
        self.close()

    def start_show(self):
        if self.isHidden():
            self.show()
            self.adjustSize()

    def display_text(self, 
                     icon: QIcon | QPixmap,
                     text: str, 
                     timeout: int = 1500, 
                     icon_size: QSize = None):

        if isinstance(icon, QIcon):
            self.icon.setPixmap(icon.pixmap(icon_size 
                                            if icon_size is not None 
                                            else QSize(32, 32)))
        else:
            self.icon.setPixmap(icon)

        self.text.setText(str(text))

        QTimer.singleShot(0, self.start_show)
        QTimer.singleShot(timeout, self.start_closeing)

    def mousePressEvent(self, event) -> None:
        if self._close_on_mouse_press:
            self.start_closeing()


class SilentTextDialog(WidgetBase, _SilentTextDialog):
    def __init__(self, 
                 uid: str = None, 
                 signals: SignalTypeVar = None, 
                 events: EventTypeVar = None, 
                 close_on_mouse_press: bool = True,
                 **kwargs):
        _SilentTextDialog.__init__(self, 
                                   close_on_mouse_press=close_on_mouse_press)
        super().__init__(uid, signals, events, **kwargs)

    def display(self, 
                icon: QIcon | QPixmap,
                text: str,
                timeout: int = 1500,
                icon_size: QSize = None):
        self.display_text(icon, text, timeout, icon_size)

    def run(self, 
            icon: QIcon | QPixmap,
            text: str, 
            timeout: int = 1500,
            icon_size: QSize = None):
        self.display_text(icon, text, timeout, icon_size)


def main():
    app = QApplication(argv)
    window = _SilentTextDialog()
    window.display_text(QIcon.fromTheme("firefox"), "Test Display Silent Text is Ok! slkdfjskjfkls jdflksjdflksdjf")
    window.show()
    exit(app.exec())

if __name__ == "__main__":
    main()



# Qtica
__version__ = "0.1.4"
__version_info__ = (0, 1, 4, "", "")

# Python
__minimum_python_version__ = (3, 7)
__maximum_python_version__ = (3, 11)

# PySide6
__pyside_version__ = "6.5.0"

# base classes
from .core import (
    AbstractBase, 
    ObjectBase, 
    WidgetBase, 
    Return,
    NoneCheck,

    # declarative
    BehaviorDeclarative, 
    WidgetDeclarative, 
    ObjectDeclarative,
    DuplicateKeyError,
    TrackingDeclarative,

    # platforms
    AndroidInit,
    IOSInit,
    LinuxInit,
    MacOSInit,
    WindowsInit,

    Api,
    QStyleSheet
)


#!/usr/bin/python3
from sys import argv

from PySide6.QtGui import QFont, QMouseEvent
from PySide6.QtWidgets import QGridLayout, QApplication, QLabel, QDialog, QWidget
from PySide6.QtCore import QTimer, Qt

from ...enums.events import EventTypeVar
from ...enums.signals import SignalTypeVar
from ...core.base import WidgetBase


class _DisplayLargText(QDialog):
    def __init__(self,
                 parent = None,
                 close_on_mouse_press: bool = True) -> None:
        QDialog.__init__(self, parent)

        self._close_on_mouse_press = close_on_mouse_press

        self.setWindowFlags(
            self.windowFlags() 
            | Qt.WindowType.FramelessWindowHint)

        self.setAttribute(Qt.WidgetAttribute.WA_TranslucentBackground, True)
        self.setAttribute(Qt.WidgetAttribute.WA_AlwaysStackOnTop, True)
        self.setWindowOpacity(0.9)

        self.setStyleSheet("""
        QDialog, QWidget{
            padding: 8px;
            border-radius: 8px;
            background-color: #0d0d0d;
            color: #ffffff;
        }
        """)

        self.form = QWidget(self)
        self.grid_layout1 = QGridLayout(self)
        self.grid_layout2 = QGridLayout(self.form)
        self.text = QLabel(self.form)

        self.font = QFont()
        self.font.setBold(True)

        self.text.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.text.setWordWrap(True)

        self.grid_layout2.addWidget(self.text, 0, 0, 0, 0, Qt.AlignmentFlag.AlignCenter)

        self.grid_layout1.addWidget(self.form)
        self.form.setLayout(self.grid_layout2)

        self.setLayout(self.grid_layout1)

    def start_closeing(self):
        self.close()

    def start_show(self):
        if self.isHidden():
            self.show()
            self.adjustSize()

    def display_text(self, 
                     text: str, 
                     font_size: int = 50, 
                     timeout: int = 2000) -> None:

        self.text.setText(text)
        self.font.setPixelSize(font_size)
        self.text.setFont(self.font)

        QTimer.singleShot(0, self.start_show)
        QTimer.singleShot(timeout, self.start_closeing)

    def mousePressEvent(self, event: QMouseEvent) -> None:
        if self._close_on_mouse_press:
            self.start_closeing()
        return


class LargTextDialog(WidgetBase, _DisplayLargText):
    def __init__(self,
                 uid: str = None, 
                 signals: SignalTypeVar = None, 
                 events: EventTypeVar = None,
                 close_on_mouse_press: bool = True,
                 **kwargs):
        _DisplayLargText.__init__(self, 
                                  close_on_mouse_press=close_on_mouse_press)
        super().__init__(uid, signals, events, **kwargs)

    def display(self, 
                text: str,
                font_size: int = 50,
                timeout: int = 2000) -> None:

        self.display_text(text, font_size, timeout)

    def run(self,
            text: str,
            font_size: int = 50,
            timeout: int = 2000) -> None:

        self.display_text(text, font_size, timeout)


def main():
    app = QApplication(argv)
    window = _DisplayLargText()
    window.display_text("Test Display Larg Text is Ok!")
    window.show()
    exit(app.exec())

if __name__ == "__main__":
    main()



#!/usr/bin/python3

from typing import Union
from PySide6.QtWidgets import QFrame, QGraphicsEffect
from PySide6.QtGui import QIcon, QIconEngine, QImage, QPaintEvent, QPainter, QPixmap
from PySide6.QtCore import Qt
from ..enums._abs_icons import AbstractIcons

from ..core.qstyle_sheet import QStyleSheet
from ..enums.events import EventTypeVar
from ..enums.signals import SignalTypeVar
from ..core.base import WidgetBase
from ..tools.icon import Icon


class IconWidget(WidgetBase, QFrame):
    def __init__(self,
        icon: Union[str,
                    QIcon,
                    QIconEngine,
                    QPixmap,
                    QImage,
                    AbstractIcons],
        alignment: Qt.AlignmentFlag = Qt.AlignmentFlag.AlignCenter,
        mode: QIcon.Mode = QIcon.Mode.Active,
        state: QIcon.State = QIcon.State.On,
        uid: str = None, 
        signals: SignalTypeVar = None, 
        events: EventTypeVar = None, 
        qss: str | dict | QStyleSheet = None, 
        attrs: list[Qt.WidgetAttribute] | dict[Qt.WidgetAttribute, bool] = None, 
        flags: list[Qt.WindowType] | dict[Qt.WindowType, bool] = None, 
        effect: QGraphicsEffect = None, 
        **kwargs):
        QFrame.__init__(self)
        super().__init__(uid, signals, events, qss, attrs, flags, effect, **kwargs)

        self.setFrameShape(QFrame.Shape.NoFrame)
        self.setFrameShadow(QFrame.Shadow.Raised)

        self._icon = Icon(icon)
        self._alignment = alignment
        self._mode = mode
        self._state = state

    def paintEvent(self, event: QPaintEvent) -> None:
        super().paintEvent(event)

        painter = QPainter(self)
        painter.setRenderHints(QPainter.RenderHint.Antialiasing 
                               | QPainter.RenderHint.SmoothPixmapTransform
                               | QPainter.RenderHint.TextAntialiasing)

        self._icon.paint(painter, 
                         self.rect(),
                         self._alignment, 
                         self._mode, 
                         self._state)


from PySide6.QtCore import Qt, Signal
from PySide6.QtWidgets import QPushButton
from ..enums.events import EventTypeVar
from ..enums.signals import SignalTypeVar
from ..core.base import WidgetBase


class PushButton(WidgetBase, QPushButton):
    long_press = Signal()
    long_repeat = Signal()
    long_release = Signal()
    long_click = Signal()

    def __init__(self, 
                 enable_long_press: bool = False,
                 uid: str = None, 
                 signals: SignalTypeVar = None, 
                 events: EventTypeVar = None, 
                 qss: str | dict = None, 
                 attrs: list[Qt.WidgetAttribute] | dict[Qt.WidgetAttribute, bool] = None, 
                 flags: list[Qt.WindowType] | dict[Qt.WindowType, bool] = None, 
                 **kwargs):
        QPushButton.__init__(self)
        super().__init__(uid, signals, events, qss, attrs, flags, **kwargs)

        self._state = 0

        if enable_long_press:
            self.setAutoRepeat(True)
            self.setAutoRepeatDelay(1000)
            self.setAutoRepeatInterval(200)

        self.clicked.connect(self._handleLongClicked)

    def _handleLongClicked(self):
        if self.isDown():
            if self._state == 0:
                self._state = 1
                self.setAutoRepeatInterval(50)
                self.long_press.emit()
            else:
                self.long_repeat.emit()

        elif self._state == 1:
            self._state = 0
            self.setAutoRepeatInterval(200)
            self.long_release.emit()

        else:
            self.long_click.emit()


# coding:utf-8

from PySide6.QtWidgets import QLabel
from PySide6.QtCore import QPoint, QRect, QSize, Qt, Signal
from PySide6.QtGui import QFont, QFontMetrics, QResizeEvent, QTextLayout

from ..enums.events import EventTypeVar
from ..enums.signals import SignalTypeVar
from ..core.base import WidgetBase


class _ElidingLabel(QLabel):
    """A QLabel variant that will elide text (add '…') to fit width.

    ElidingLabel()
    ElidingLabel(parent: Optional[QWidget], f: Qt.WindowFlags = ...)
    ElidingLabel(text: str, parent: Optional[QWidget] = None, f: Qt.WindowFlags = ...)

    For a multiline eliding label, use `setWordWrap(True)`.  In this case, text
    will wrap to fit the width, and only the last line will be elided.
    When `wordWrap()` is True, `sizeHint()` will return the size required to fit
    the full text.
    """

    textChanged = Signal(str)

    def __init__(self, text: str = None) -> None:
        self._elide_mode = Qt.TextElideMode.ElideRight
        super().__init__()
        
        self._silent_set_text(text if text else "")

    def _silent_set_text(self, text: str):
        self.blockSignals(True)
        self.setText(text)
        self.blockSignals(False)

    # New Public methods

    def elideMode(self) -> Qt.TextElideMode:
        """The current Qt.TextElideMode."""
        return self._elide_mode

    def setElideMode(self, mode: Qt.TextElideMode):
        """Set the elide mode to a Qt.TextElideMode."""
        self._elide_mode = Qt.TextElideMode(mode)
        super().setText(self._elidedText())

    @staticmethod
    def wrapText(text, width, font=None) -> list[str]:
        """Returns `text`, split as it would be wrapped for `width`, given `font`.

        Static method.
        """
        tl = QTextLayout(text, font or QFont())
        tl.beginLayout()
        lines = []
        while True:
            ln = tl.createLine()
            if not ln.isValid():
                break
            ln.setLineWidth(width)
            start = ln.textStart()
            lines.append(text[start : start + ln.textLength()])
        tl.endLayout()
        return lines

    # Reimplemented QT methods

    def text(self) -> str:
        """This property holds the label's text.

        If no text has been set this will return an empty string.
        """
        return self._text

    def setText(self, txt: str):
        """Set the label's text.

        Setting the text clears any previous content.
        NOTE: we set the QLabel private text to the elided version
        """
        self._text = txt
        super().setText(self._elidedText())
        self.textChanged.emit(txt)

    def resizeEvent(self, ev: QResizeEvent) -> None:
        ev.accept()
        super().setText(self._elidedText())

    def setWordWrap(self, wrap: bool) -> None:
        super().setWordWrap(wrap)
        super().setText(self._elidedText())

    def sizeHint(self) -> QSize:
        if not self.wordWrap():
            return super().sizeHint()
        fm = QFontMetrics(self.font())
        flags = int(self.alignment() | Qt.TextFlag.TextWordWrap)
        r = fm.boundingRect(QRect(QPoint(0, 0), self.size()), flags, self._text)
        print(self.width(), r.height())
        return QSize(self.width(), r.height())

    # private implementation methods
    def _elidedText(self) -> str:
        """Return `self._text` elided to `width`"""
        fm = QFontMetrics(self.font())
        # the 2 is a magic number that prevents the ellipses from going missing
        # in certain cases (?)
        width = self.width() - 2
        if not self.wordWrap():
            return fm.elidedText(self._text, self._elide_mode, width)

        # get number of lines we can fit without eliding
        nlines = self.height() // fm.height() - 1
        # get the last line (elided)
        text = self._wrappedText()
        last_line = fm.elidedText("".join(text[nlines:]), self._elide_mode, width)
        # join them
        return "".join(text[:nlines] + [last_line])

    def _wrappedText(self) -> list[str]:
        return _ElidingLabel.wrapText(self._text, self.width(), self.font())


class ElidingLabel(WidgetBase, _ElidingLabel):
    def __init__(self, 
                 text: str = None,
                 elide_mode: Qt.TextElideMode = None,
                 uid: str = None, 
                 signals: SignalTypeVar = None, 
                 events: EventTypeVar = None,
                 **kwargs):
        _ElidingLabel.__init__(self, text)
        super().__init__(uid, signals, events, **kwargs)

        if elide_mode is not None:
            self.setElideMode(elide_mode)


from PySide6.QtCore import QUrl
from PySide6.QtQuick import QQuickView

from ...enums.events import EventTypeVar
from ...enums.signals import SignalTypeVar
from ...core.base import ObjectBase


class QuickView(ObjectBase, QQuickView):
    def __init__(self,
                 file: QUrl | str,
                 uid: str = None, 
                 signals: SignalTypeVar = None, 
                 events: EventTypeVar = None, 
                 **kwargs):
        QQuickView.__init__(self, file)
        super().__init__(uid, signals, events, **kwargs)

        self.setResizeMode(QQuickView.ResizeMode.SizeRootObjectToView)


from PySide6.QtCore import QByteArray, QSize, Qt
from PySide6.QtGui import QColor, QMovie
from PySide6.QtWidgets import QLabel
from ..enums.events import EventTypeVar
from ..enums.signals import SignalTypeVar
from ..core.base import WidgetBase
import os


class MovieView(WidgetBase, QLabel):
    def __init__(self,
                 filename: str = None,
                 bg_color: QColor = None,
                 scale_size: QSize = None,
                 speed: int = None,
                 running: bool = False,
                 uid: str = None, 
                 signals: SignalTypeVar = None, 
                 events: EventTypeVar = None,
                 qss: str | dict = None,
                 attrs: list[Qt.WidgetAttribute] | dict[Qt.WidgetAttribute, bool] = None,
                 flags: list[Qt.WindowType] | dict[Qt.WindowType, bool] = None,
                 **kwargs):
        QLabel.__init__(self)
        super().__init__(uid, signals, events, qss, attrs, flags, **kwargs)

        if (QByteArray(bytes(os.path.basename(filename).split(".")[-1].lower(), "utf-8")) 
            not in QMovie.supportedFormats()):
            raise ValueError("invalid filename, unsupported format!")

        self.setMouseTracking(True)
        self.setAlignment(Qt.AlignmentFlag.AlignCenter)

        self._movie = QMovie(self)

        if filename is not None:
            self._movie.setFileName(filename)

        if bg_color is not None:
            self._movie.setBackgroundColor(bg_color)

        if scale_size is not None:
            self._movie.setScaledSize(scale_size)

        if speed is not None:
            self._movie.setSpeed(speed)

        if running:
            self._movie.start()

        self.setMovie(self._movie)



#!/usr/bin/python3

from PySide6.QtCore import QFile
from typing import Union
from string import Template

import os


class _QssTemplate(Template):
    delimiter = "--"


class QStyleSheet:
    '''
    Label(
        uid="label",
        qss=QStyleSheet(
            qss={
                "background-color": "--themeColor",
                "color": "white",
                "font-size": "24px"
            },
            vars={
                "themeColor": "blue"
            }
        )
    )
    
    :param: qss: 
        - {"background-color": "white"}
        - ":/path/to/resoucre"
        - "/path/to/local/file"
        - QWidget { background-color: white;}
    '''

    def __init__(self, 
                 qss: Union[str, dict],
                 vars: dict = None):

        self._qss = qss
        self._vars = vars
        self._temp = _QssTemplate("")
        self._parent = None

    def _set_parent(self, parent: object):
        self._parent = parent

    def _get_qss_from_str(self, qss: str) -> str:
        if qss.startswith(":"):
            file = QFile(qss)
            file.open(QFile.OpenModeFlag.ReadOnly)
            data = str(file.readAll(), "utf-8")
            file.close()
            return data

        elif os.path.exists(qss):
            with open(qss, "r") as fr:
                return fr.read()
        return qss

    def _get_qss_from_dict(self, qss: dict) -> str:
        style_sheet = ""
        _obj_style = ""

        for k, v in qss.items():
            if isinstance(v, dict):
                style_sheet += "%s {\n" % k
                for sk, sv in v.items():
                    style_sheet += f"{sk}: {sv};\n"
                style_sheet += "}\n"
            else:
                _obj_style += f"{k}: {v};\n"

        if (obj_name := self._parent.objectName()):
            style_sheet += "#%s {\n" % obj_name
            style_sheet += _obj_style
            style_sheet += "}"

        elif not style_sheet:
            style_sheet = _obj_style

        return style_sheet

    def _set_qss(self, qss: Union[str, dict]) -> None:
        _style = (self._get_qss_from_str(qss)
                  if isinstance(qss, str) 
                  else self._get_qss_from_dict(qss))

        if self._vars:
            self._temp.template = _style
            self._temp.safe_substitute(_style)
            _style = self._temp.safe_substitute(self._vars)

        self._parent.setStyleSheet(_style)

    def update_vars(self, vars: dict) -> None:
        self._vars = vars
        self._set_qss(self._qss)

    def update_qss(self,
                   qss: Union[str, dict],
                   *,
                   save: bool = False) -> None:

        if (isinstance(qss, dict) 
            and isinstance(self._qss, dict)):
            _qss = self._qss.copy()
            _qss.update(qss)
            self._qss = qss if save else self._qss
            _style = self._get_qss_from_dict(_qss)
        else:
            _style = self._get_qss_from_str(qss)
            self._qss = qss if save else self._qss

        if self._vars:
            self._temp.template = _style
            self._temp.safe_substitute(_style)
            _style = self._temp.safe_substitute(self._vars)

        self._parent.setStyleSheet(_style)

    def update(self, qss: Union[str, dict], vars: dict = None, *, save: bool = False) -> None:
        if vars is not None:
            self.update_vars(vars)
        self.update_qss(qss, save=save)


from PySide6.QtWidgets import QFrame, QScrollArea, QWidget
from PySide6.QtCore import Qt
from ..enums.events import EventTypeVar
from ..enums.signals import SignalTypeVar
from ..core.base import WidgetBase


class ScrollArea(WidgetBase, QScrollArea):
    def __init__(self, 
                 child: QWidget = None,
                 uid: str = None,
                 signals: SignalTypeVar = None, 
                 events: EventTypeVar = None, 
                 qss: str | dict = None, 
                 attrs: list[Qt.WidgetAttribute] | dict[Qt.WidgetAttribute, bool] = None, 
                 flags: list[Qt.WindowType] | dict[Qt.WindowType, bool] = None, 
                 **kwargs):
        QScrollArea.__init__(self)
        super().__init__(uid, signals, events, qss, attrs, flags, **kwargs)

        self.setWidgetResizable(True)
        self.setUpdatesEnabled(True)

        self.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.setSizeAdjustPolicy(QScrollArea.SizeAdjustPolicy.AdjustToContents)

        self.setFrameShadow(QFrame.Shadow.Plain)
        self.setFrameShape(QFrame.Shape.NoFrame)

        if child is not None:
            self.setWidget(child)


from PySide6.QtCore import Qt, QEvent
from PySide6.QtGui import QPainter, QColor
from PySide6.QtWidgets import QGridLayout, QWidget

from darkdetect import isDark as isDarkTheme

from ...enums.events import EventTypeVar
from ...enums.signals import SignalTypeVar
from ...enums import TeachingTipTailPositions
from .tails import TeachingTipManager


class _TeachingTip(QWidget):
    def __init__(self, 
                 view: QWidget,
                 target: QWidget,
                 tail_position: TeachingTipTailPositions = TeachingTipTailPositions.bottom,
                 delete_on_close: bool = False,
                 radius: int = 6,
                 parent: QWidget = None
                ) -> None:
        super().__init__(parent=parent)

        self.view = view
        self.target = target
        self.delete_on_close = delete_on_close
        self.manager = TeachingTipManager.make(tail_position)
        self.manager.radius = radius

        self._layout = QGridLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(0)
        self.manager.doLayout(self)

        self._layout.addWidget(view)

        # set style
        self.setAttribute(Qt.WidgetAttribute.WA_TranslucentBackground)
        self.setAttribute(Qt.WidgetAttribute.WA_StyledBackground)
        self.setWindowFlags(Qt.WindowType.Tool | Qt.WindowType.FramelessWindowHint)

        if parent and parent.window():
            parent.window().installEventFilter(self)

        if self.target is not None:
            self.target.resizeEvent = lambda e: self.update_pos()
            self.target.moveEvent = lambda e: self.update_pos()
            self.target.window().moveEvent = lambda e: self.update_pos()

    def tooltip_pos(self):
        return self.manager.position(self)

    def set_target(self, target: QWidget):
        self.target = target
        self.target.resizeEvent = lambda e: self.update_pos()
        self.target.moveEvent = lambda e: self.update_pos()
        self.target.window().moveEvent = lambda e: self.update_pos()

    def update_pos(self):
        self.move(self.manager.position(self))

    def showEvent(self, e):
        self.update_pos()
        super().showEvent(e)

    def closeEvent(self, e):
        if self.delete_on_close:
            self.deleteLater()
        super().closeEvent(e)

    def eventFilter(self, obj, e: QEvent):
        if self.parent() and obj is self.parent().window():
            if e.type() in [QEvent.Type.Resize, 
                            QEvent.Type.WindowStateChange, 
                            QEvent.Type.Move]:
                self.move(self.manager.position(self))
        return super().eventFilter(obj, e)

    def paintEvent(self, e, scale: float = None):
        super().paintEvent(e)

        painter = QPainter(self)
        painter.setRenderHints(QPainter.RenderHint.Antialiasing)

        painter.setBrush(
            QColor(40, 40, 40) 
            if isDarkTheme() 
            else QColor(248, 248, 248))
        
        painter.setPen(
            QColor(23, 23, 23) 
            if isDarkTheme() 
            else QColor(195, 195, 195))

        self.manager.draw(self, painter)


from ...core.base import WidgetBase

class TeachingTip(WidgetBase, _TeachingTip):
    def __init__(self, 
                 child: QWidget,
                 target: QWidget,
                 tail_position: TeachingTipTailPositions = TeachingTipTailPositions.bottom,
                 delete_on_close: bool = False,
                 radius: int = 6,
                 uid: str = None, 
                 signals: SignalTypeVar = None, 
                 events: EventTypeVar = None, 
                 qss: str | dict = None, 
                 attrs: list[Qt.WidgetAttribute] | dict[Qt.WidgetAttribute, bool] = None, 
                 flags: list[Qt.WindowType] | dict[Qt.WindowType, bool] = None,
                 **kwargs):
        _TeachingTip.__init__(self, child, target, tail_position, delete_on_close, radius)
        super().__init__(uid, signals, events, qss, attrs, flags, **kwargs)


from PySide6.QtWidgets import QSizePolicy, QWidget
from ..core.base import BehaviorDeclarative


class SizePolicy(BehaviorDeclarative):
    def __init__(self, 
                 child: QWidget, 
                 horizontal: QSizePolicy.Policy = None,
                 vertical: QSizePolicy.Policy = None,
                 stretch: tuple[int, int] = None,
                 type: QSizePolicy.ControlType = None) -> QWidget:

        self._size_policy = QSizePolicy()

        if type is not None:
            self._size_policy.setControlType(type)

        if horizontal is not None:
            self._size_policy.setHorizontalPolicy(horizontal)

        if vertical is not None:
            self._size_policy.setVerticalPolicy(vertical)

        self._size_policy.setHorizontalStretch(stretch[0] if stretch is not None else 0)
        self._size_policy.setVerticalStretch(stretch[-1] if stretch is not None else 0)

        self._size_policy.setHeightForWidth(child.sizePolicy().hasHeightForWidth())
        self._size_policy.setWidthForHeight(child.sizePolicy().hasWidthForHeight())

        child.setSizePolicy(self._size_policy)

        # sizePolicy.setRetainSizeWhenHidden(True)

        return child

    def build(self):
        self._size_policy


from PySide6.QtCore import Qt
from PySide6.QtWidgets import QLineEdit
from ..enums.events import EventTypeVar
from ..enums.signals import SignalTypeVar
from ..core.base import WidgetBase


class LineEdit(WidgetBase, QLineEdit):
    def __init__(self, 
                 uid: str = None, 
                 signals: SignalTypeVar = None, 
                 events: EventTypeVar = None, 
                 qss: str | dict = None, 
                 attrs: list[Qt.WidgetAttribute] | dict[Qt.WidgetAttribute, bool] = None, 
                 flags: list[Qt.WindowType] | dict[Qt.WindowType, bool] = None, 
                 **kwargs):
        QLineEdit.__init__(self)
        super().__init__(uid, signals, events, qss, attrs, flags, **kwargs)


from .base import (
    AbstractBase,
    ObjectBase, 
    WidgetBase, 
    Return,
    NoneCheck,

    # declarative
    ObjectDeclarative,
    WidgetDeclarative,
    DuplicateKeyError,
    BehaviorDeclarative,
    TrackingDeclarative
)

from .init import (
    AndroidInit,
    IOSInit,
    LinuxInit,
    MacOSInit,
    WindowsInit
)
from .api import Api
from .qstyle_sheet import QStyleSheet


from PySide6.QtCore import Signal
from PySide6.QtGui import QCloseEvent, QShowEvent
from PySide6.QtWidgets import (
    QMainWindow, 
    QWidget, 
    QToolBar,
    QStatusBar,
    QDockWidget,
    QStackedWidget,
    QLayout
)
from ...enums.events import EventTypeVar
from ...enums.signals import SignalTypeVar
from ...core.base import WidgetBase
from typing import Dict, Union


class Routes:
    def __init__(self, 
                 stacked_widget: QStackedWidget = None,
                 index: str = "/") -> None:
        
        self._stacked_widget = stacked_widget
        self._index = index
        self._routes: dict = {}

    def add(self, route: str, widget: QWidget) -> None:
        self._routes[route.strip()] = self._stacked_widget.addWidget(widget)

    def push(self, route: str) -> None:
        self._stacked_widget.setCurrentIndex(self._routes[route])

    def pop(self) -> None:
        self._stacked_widget.setCurrentIndex(self._routes[self._index])

    def remove(self, route: str) -> None:
        index = self._routes.pop(route)
        self._stacked_widget.removeWidget(self.widget(index))

    def stackedWidget(self) -> QStackedWidget:
        return self._stacked_widget


class RoutingWindow(WidgetBase, QMainWindow):
    startup_changed = Signal()

    def __init__(self, 
                 uid: str = None, 
                 signals: SignalTypeVar = None,
                 events: EventTypeVar = None, 
                 index: str = "/",
                 routes: Dict[str, Union[QWidget, QLayout]] = None,
                 stacked_widget: QStackedWidget = None,
                 tool_bar: QToolBar = None,
                 status_bar: QStatusBar = None,
                 dock_widget: QDockWidget = None,
                 **kwargs):
        QMainWindow.__init__(self)
        super().__init__(uid, signals, events, **kwargs)

        self.__is_startup = False
        stacked_widget = (stacked_widget 
                          if stacked_widget is not None 
                          else QStackedWidget())

        stacked_widget.setParent(self)
        self._routes = Routes(stacked_widget, index)

        self.setAnimated(True)
        self.setUpdatesEnabled(True)
        self.setDocumentMode(True)

        if routes is not None:
            self._set_routes(routes)

        if tool_bar is not None:
            self._set_tool_bar(tool_bar)
    
        if status_bar is not None:
            self._set_status_bar(status_bar)

        if dock_widget is not None:
            self._set_dock_widget(dock_widget)

    def closeEvent(self, event: QCloseEvent) -> None:
        self.__is_startup = False
        return super().closeEvent(event)

    def showEvent(self, event: QShowEvent) -> None:
        if not self.__is_startup:
            self.__is_startup = True
            self.startup_changed.emit()

        return super().showEvent(event)

    def _set_routes(self, routes: Dict[str, QWidget]) -> None:
        for route, child in routes.items():
            if isinstance(child, QLayout):
                widget = QWidget(self)
                widget.setLayout(child)
                self._routes.add(route, widget)
            else:
                self._routes.add(route, child)

        self.setCentralWidget(self._routes.stackedWidget())

    def _set_tool_bar(self, tool_bar: QToolBar) -> None:
        self.addToolBar(tool_bar)

    def _set_status_bar(self, status_bar: QStatusBar) -> None:
        self.setStatusBar(status_bar)

    def _set_dock_widget(self, dock_widget: QDockWidget) -> None:
        self.addDockWidget(dock_widget)

    def push(self, route: str) -> None:
        self.route.push(route)

    @property
    def route(self) -> Routes:
        return self._routes